r"""Black-Scholes-Merton options pricing formulae using dividend yield.

* Stock price $ S $,
* Strike price $ K $,
* Risk-free rate $ r $,
* Annual dividend yield $ q $,
* Time to maturity $ \tau = T - t $
* Volatility $ \sigma $.

where:

$$
d_1 &= \frac{\ln(S/K) + \left(r - q + \frac{1}{2}\sigma^2\right)\tau}{\sigma\sqrt{\tau}}
$$

$$
d_2 &= \frac{\ln(S/K) + \left(r - q - \frac{1}{2}\sigma^2\right)\tau}{\sigma\sqrt{\tau}} = d_1 - \sigma\sqrt{\tau}
$$

$$
\varphi(x) &= \frac{1}{\sqrt{2\pi}} e^{-\frac{1}{2} x^2}
$$

$$
\Phi(x) &= \frac{1}{\sqrt{2\pi}} \int_{-\infty}^x e^{-\frac{1}{2} y^2} \,dy = 1 - \frac{1}{\sqrt{2\pi}} \int_x^\infty e^{-\frac{1}{2} y^2} \,dy
$$
"""

from math import exp, log, pi, sqrt
from statistics import NormalDist
from typing import Literal

from ..implied_volatility import solve_ivol
from ..numeric_greeks.with_dividend_yield import NumericGreeks


norm = NormalDist()
cdf = norm.cdf
pdf = norm.pdf
inv_cdf = norm.inv_cdf


def price(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    r"""The fair value of a European option, using Black-Scholes-Merton.

    Call price: $Se^{-q \tau}\Phi(d_1) - e^{-r \tau} K\Phi(d_2)$

    Put price: $e^{-r \tau} K\Phi(-d_2) -  Se^{-q \tau}\Phi(-d_1)$

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        q (float): The dividend yield.
        v (float): The volatility of the asset.

    Returns:
        float: The price of the options.
    """

    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)

    F = S * exp((r - q) * T)
    if is_call:
        return exp(-r * T) * (F * cdf(d1) - K * cdf(d2))
    else:
        return exp(-r * T) * (K * cdf(-d2) - F * cdf(-d1))


def ivol(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        p: float,
        *,
        max_iterations: int = 20,
        epsilon=1e-8
) -> float:
    """Calculate the volatility of an option that is implied by the price.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        q (float): The dividend yield.
        p (float): The option price.
        max_iterations (int, Optional): The maximum number of iterations before
            a price is returned. Defaults to 20.
        epsilon (float, Optional): The largest acceptable error. Defaults to 1e-8.

    Returns:
        float: The implied volatility.
    """
    return solve_ivol(
        p,
        lambda v: price(is_call, S, K, T, r, q, v),
        max_iterations=max_iterations,
        epsilon=epsilon
    )


def make_numeric_greeks(is_call: bool) -> NumericGreeks:
    def evaluate(S: float, K: float, T: float, r: float, q: float, v: float) -> float:
        return price(is_call, S, K, T, r, q, v)
    return NumericGreeks(evaluate)


def delta(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    r"""The sensitivity of the open to a change in the asset price.

    Call $\Delta$  $e^{-q \tau} \Phi(d_1)$

    Put $\Delta$ $-e^{-q \tau} \Phi(-d_1)$

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        q (float): The dividend yield.
        v (float): The volatility of the asset.

    Returns:
        float: the delta.
    """
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))

    if is_call:
        return exp(-q * T) * cdf(d1)
    else:
        return -exp(-q * T) * cdf(-d1)


def gamma(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    r"""The second derivative to the change in the asset price.

    $$
    \Gamma $ $ e^{-q \tau} \frac{\varphi(d_1)}{S\sigma\sqrt{\tau}} = K e^{-r \tau} \frac{\varphi(d_2)}{S^2\sigma\sqrt{\tau}}
    $$

    Args:
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        q (float): The dividend yield.
        v (float): The volatility of the asset.

    Returns:
        float: The gamma.
    """
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))

    return exp(-q * T) * pdf(d1) / (S * v * sqrt(T))


def theta(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    r"""The theta or time decay of the value of the option.

    $$
    Call \Theta $ $ - e^{-q \tau} \frac{S \varphi(d_1) \sigma}{2 \sqrt{\tau}} - rKe^{-r \tau}\Phi(d_2) + qSe^{-q \tau}\Phi(d_1)
    $$

    $$
    Put \Theta $ $ - e^{-q \tau}\frac{S \varphi(d_1) \sigma}{2 \sqrt{\tau}} + rKe^{-r \tau}\Phi(-d_2) - qSe^{-q \tau}\Phi(-d_1)
    $$

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The asset price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        q (float): The dividend yield.
        v (float): The asset volatility.

    Returns:
        float: The theta.
    """
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)

    if is_call:
        p1 = -S * exp(-q * T) * pdf(d1) * v / (2 * sqrt(T))
        p2 = -q * S * exp(-q * T) * cdf(d1)
        p3 = r * K * exp(-r * T) * cdf(d2)
        return p1 - p2 - p3
    else:
        p1 = -S * exp(-q * T) * pdf(d1) * v / (2 * sqrt(T))
        p2 = -q * S * exp(-q * T) * cdf(-d1)
        p3 = r * K * exp(-r * T) * cdf(-d2)
        return p1 + p2 + p3


def vega(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    r"""The sensitivity of the options price or a change in the asset volatility.

    $$
    \mathcal{V} $ is $ S e^{-q \tau} \varphi(d_1) \sqrt{\tau} = K e^{-r \tau} \varphi(d_2) \sqrt{\tau}
    $$

    Args:
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        q (float): The dividend yield.
        v (float): The volatility of the asset.

    Returns:
        float: The vega
    """
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    return S * exp(-q * T) * pdf(d1) * sqrt(T)


def rho(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    r"""The sensitivity of the option price to the risk free rate.

    Call $ \rho $ is $ K \tau e^{-r \tau}\Phi(d_2) $

    Put $ \rho $ is $ -K \tau e^{-r \tau}\Phi(-d_2) $

    Useful for all options except futures options which should use
    futures_rho.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The asset price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        q (float): The dividend yield.
        v (float): The asset volatility.

    Returns:
        float: The rho.
    """
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    if is_call:
        return K * T * exp(-r * T) * cdf(d2)
    else:
        return -K * T * exp(-r * T) * cdf(-d2)


def carry(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    """Sensitivity to the cost of carry.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The asset price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        q (float): The dividend yield.
        v (float): The asset volatility.

    Returns:
        float: The carry.
    """
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    if is_call:
        return T * S * exp(-q * T) * cdf(d1)
    else:
        return -T * S * exp(-q * T) * cdf(-d1)


def elasticity(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    """The option elasticity.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The asset price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        q (float): The dividend yield.
        v (float): The asset volatility.

    Returns:
        float: The elasticity.
    """
    return (
        delta(is_call, S, K, T, r, q, v) * S /
        price(is_call, S, K, T, r, q, v)
    )


def gammap(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    return S * gamma(S, K, T, r, q, v) / 100


def vegap(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    return v / 10 * vega(S, K, T, r, q, v)


def forward_delta(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:

    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))

    if is_call:
        return exp(-r * T) * cdf(d1)
    else:
        return exp(-r * T) * (cdf(d1) - 1)


def vanna(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    r"""The sensitivity to the spot price and volatility.

    $$
    -e^{-q \tau} \varphi(d_1) \frac{d_2}{\sigma} \, = \frac{\mathcal{V}}{S}\left[1 - \frac{d_1}{\sigma\sqrt{\tau}} \right]
    $$

    Also known as DdeltaDvol.

    Args:
        S (float): The asset price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        q (float): The dividend yield.
        v (float): The asset volatility.

    Returns:
        float: The vanna.
    """
    # Also known as DdeltaDvol.
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return -exp(-q * T) * d2 / v * pdf(d1)


def ddelta_dvol_dvol(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    # Also known as DVannaDvol
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return vanna(S, K, T, r, q, v) * 1 / v * (d1 * d2 - d1 / d2 - 1)


def charm(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    # Also known as DdeltaDtime

    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)

    if is_call:
        return (
            q * exp(-q * T) * cdf(d1)
            - exp(-q * T) * pdf(d1) * (
                2 * (r - q) * T - d2 * v * sqrt(T)
            ) / (2 * T * v * sqrt(T)))
    else:
        return (
            -q * exp(-q * T) * cdf(-d1)
            - exp(-q * T) * pdf(d1) * (
                2 * (r - q) * T - d2 * v * sqrt(T)
            ) / (2 * T * v * sqrt(T)))


def saddle_gamma(
        K: float,
        q: float,
        v: float
) -> float:
    return sqrt(exp(1) / pi) * sqrt((2 * -q) / v ** 2 + 1) / K


def dgamma_dspot(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    # Also known as Speed
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    return -gamma(S, K, T, r, q, v) * (1 + d1 / (v * sqrt(T))) / S


def dgamma_dvol(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    # Also known as zomma.
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return gamma(S, K, T, r, q, v) * ((d1 * d2 - 1) / v)


def dgamma_dtime(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return gamma(S, K, T, r, q, v) * (
        q + (r - q) * d1 / (v * sqrt(T)) +
        (1 - d1 * d2) / (2 * T)
    )


def dgammap_dspot(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    # Also known as SpeedP.
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    return -gamma(S, K, T, r, q, v) * (d1) / (100 * v * sqrt(T))


def dgammap_dvol(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return S / 100 * gamma(S, K, T, r, q, v) * ((d1 * d2 - 1) / v)


def dgammap_dtime(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return (
        gammap(S, K, T, r, q, v) *
        (q + (r - q) * d1 / (v * sqrt(T)) + (1 - d1 * d2) / (2 * T))
    )


def dvega_dtime(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return (
        vega(S, K, T, r, q, v) *
        (q + (r - q) * d1 / (v * sqrt(T)) - (1 + d1 * d2) / (2 * T))
    )


def vomma(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    # Also known as DvegaDvol
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return vega(S, K, T, r, q, v) * d1 * d2 / v


def dvomma_dvol(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return (
        vomma(S, K, T, r, q, v) *
        1 / v * (d1 * d2 - d1 / d2 - d2 / d1 - 1)
    )


def dvegap_dvol(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    # Also known as VommaP.
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return vegap(S, K, T, r, q, v) * d1 * d2 / v


def vega_leverage(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    return (
        vega(S, K, T, r, q, v) * v /
        price(is_call, S, K, T, r, q, v)
    )


def variance_vega(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    return S * exp(-q * T) * pdf(d1) * sqrt(T) / (2 * v)


def variance_delta(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return S * exp(-q * T) * pdf(d1) * (-d2) / (2 * v ** 2)


def variance_vomma(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return S * exp(-q * T) * sqrt(T) / (4 * v ** 3) * pdf(d1) * (d1 * d2 - 1)


def variance_ultima(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return (
        S * exp(-q * T) * sqrt(T) /
        (8 * v ** 5) * pdf(d1) *
        ((d1 * d2 - 1) *
         (d1 * d2 - 3) -
         (d1 ** 2 + d2 ** 2))
    )


def theta_driftless(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    return -S * exp(-q * T) * pdf(d1) * v / (2 * sqrt(T))


def futures_rho(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    return -T * price(is_call, S, K, T, r, 0, v)


def phi(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    # Also known as rho2.
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    if is_call:
        return -T * S * exp(-q * T) * cdf(d1)
    else:
        return T * S * exp(-q * T) * cdf(-d1)


def dzeta_dvol(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    if is_call:
        return -pdf(d2) * d1 / v
    else:
        return pdf(d2) * d1 / v


def dzeta_dtime(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    d1 = (log(S / K) + T * (r - q + v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    if is_call:
        return pdf(d2) * ((r - q) / (v * sqrt(T)) - d1 / (2 * T))
    else:
        return -pdf(d2) * ((r - q) / (v * sqrt(T)) - d1 / (2 * T))


def break_even_probability(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    # Risk neutral break even probability.
    if is_call:
        K = K + price(True, S, K, T, r, q, v) * exp(r * T)
        d2 = (log(S / K) + (r - q - v ** 2 / 2) * T) / (v * sqrt(T))
        return cdf(d2)
    else:
        K = K - price(False, S, K, T, r, q, v) * exp(r * T)
        d2 = (log(S / K) + (r - q - v ** 2 / 2) * T) / (v * sqrt(T))
        return cdf(-d2)


def strike_delta(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    d2 = (log(S / K) + (r - q - v ** 2 / 2) * T) / (v * sqrt(T))
    if is_call:
        return -exp(-r * T) * cdf(d2)
    else:
        return exp(-r * T) * cdf(-d2)


def risk_neutral_density(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    d2 = (log(S / K) + (r - q - v ** 2 / 2) * T) / (v * sqrt(T))
    return exp(-r * T) * pdf(d2) / (K * v * sqrt(T))


def gamma_from_delta(
        S: float,
        T: float,
        q: float,
        v: float,
        delta_: float,
) -> float:
    return exp(-q * T) * pdf(
        inv_cdf(exp(q * T) * abs(delta_))
    ) / (S * v * sqrt(T))


def gammap_from_delta(
        S: float,
        T: float,
        q: float,
        v: float,
        delta_: float,
) -> float:
    return S / 100 * gamma_from_delta(S, T, q, v, delta_)


def vega_from_delta(
        S: float,
        T: float,
        q: float,
        delta_: float,
) -> float:
    return S * exp(-q * T) * sqrt(T) * pdf(
        inv_cdf(exp(q * T) * abs(delta_))
    )


def vegap_from_delta(
        S: float,
        T: float,
        q: float,
        v: float,
        delta_: float,
) -> float:
    return v / 10 * vega_from_delta(S, T, q, delta_)


def strike_from_delta(
        is_call: bool,
        S: float,
        T: float,
        r: float,
        q: float,
        v: float,
        delta_: float,
) -> float:
    if is_call:
        return S * exp(
            -inv_cdf(delta_ * exp(q * T)) *
            v * sqrt(T) + (r - q + v * v / 2) * T
        )
    else:
        return S * exp(
            inv_cdf(-delta_ * exp(q * T)) *
            v * sqrt(T) + (r - q + v * v / 2) * T
        )


def in_the_money_prob_from_delta(
        is_call: bool,
        T: float,
        r: float,
        q: float,
        v: float,
        delta_: float,
) -> float:
    if is_call:
        return cdf(inv_cdf(delta_ / exp(-q * T)) - v * sqrt(T))
    else:
        return cdf(inv_cdf(-delta_ / exp(-q * T)) + v * sqrt(T))


def strike_from_in_the_money_prob(
        is_call: bool,
        S: float,
        v: float,
        T: float,
        r: float,
        q: float,
        in_the_money_prob: float,
) -> float:
    if is_call:
        return S * exp(
            -inv_cdf(in_the_money_prob) * v * sqrt(T) + (r - q - v * v / 2) * T
        )
    else:
        return S * exp(
            inv_cdf(in_the_money_prob) * v * sqrt(T) + (r - q - v * v / 2) * T
        )


def rnd_from_in_the_money_prob(
        K: float,
        T: float,
        r: float,
        v: float,
        in_the_money_prob: float,
) -> float:
    return exp(-r * T) * pdf(inv_cdf(in_the_money_prob)) / (K * v * sqrt(T))


def delta_from_in_the_money_prob(
        is_call: bool,
        T: float,
        r: float,
        q: float,
        v: float,
        in_the_money_prob: float,
) -> float:
    if is_call:
        return cdf(inv_cdf(in_the_money_prob * exp(-q * T)) - v * sqrt(T))
    else:
        return -cdf(inv_cdf(in_the_money_prob * exp(-q * T)) + v * sqrt(T))


def max_ddelta_dvol_asset(
        is_lower: bool,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float
) -> float:
    # What asset price that gives maximum DdeltaDvol

    # is_lower == True gives lower asset level that gives max DdeltaDvol
    # is_lower == False gives upper asset level that gives max DdeltaDvol

    if is_lower:
        return K * exp((q - r) * T - v * sqrt(T) * sqrt(4 + T * v ** 2) / 2)
    else:
        return K * exp((q - r) * T + v * sqrt(T) * sqrt(4 + T * v ** 2) / 2)


def max_ddelta_dvol_strike(
        is_lower: bool,
        S: float,
        T: float,
        r: float,
        q: float,
        v: float
) -> float:
    # What strike price that gives maximum DdeltaDvol

    # is_lower == True gives lower strike level that gives max DdeltaDvol
    # is_lower == False gives upper strike level that gives max DdeltaDvol

    if is_lower:
        return S * exp((r - q) * T - v * sqrt(T) * sqrt(4 + T * v * 2) / 2)
    else:
        return S * exp((r - q) * T + v * sqrt(T) * sqrt(4 + T * v ** 2) / 2)


def max_gamma_vega_at_X(
        S: float,
        r: float,
        q: float,
        T: float,
        v: float
) -> float:
    # What strike price that gives maximum gamma and vega
    return S * exp((r - q + v * v / 2) * T)


def max_gamma_at_S(
        x: float,
        r: float,
        q: float,
        T: float,
        v: float
) -> float:
    # What asset price that gives maximum gamma
    return x * exp((q - r - 3 * v * v / 2) * T)


def max_vega_at_S(
        K: float,
        r: float,
        q: float,
        T: float,
        v: float
) -> float:
    # What asset price that gives maximum vega
    return K * exp((q - r + v * v / 2) * T)


def in_the_money_probability(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
) -> float:
    d2 = (log(S / K) + (r - q - v ** 2 / 2) * T) / (v * sqrt(T))

    if is_call:
        return cdf(d2)
    else:
        return cdf(-d2)


def delta_mirror_strike(
        S: float,
        T: float,
        r: float,
        q: float,
        v: float
) -> float:
    return S * exp((r - q + v ** 2 / 2) * T)


def probability_mirror_strike(
        S: float,
        T: float,
        r: float,
        q: float,
        v: float
) -> float:
    return S * exp((r - q - v ** 2 / 2) * T)


def delta_mirror_call_put_strike(
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float
) -> float:
    return S ** 2 / K * exp((2 * (r - q) + v ** 2) * T)


def profit_loss_std(
        TypeFlag: Literal['a', 'p'],
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        q: float,
        v: float,
        NHedges: int,
) -> float:

    if TypeFlag == "a":  # in dollars
        return (
            sqrt(pi / 4) *
            vega(S, K, T, r, q, v) *
            v / sqrt(NHedges)
        )
    elif TypeFlag == "p":  # in percent
        return (
            sqrt(pi / 4) *
            vega(S, K, T, r, q, v) *
            v / sqrt(NHedges) /
            price(is_call, S, K, T, r, q, v)
        )
    else:
        raise ValueError('invalid type flag')



"""Option valuations using a trinomial tree.
"""

from math import exp, nan, sqrt
from typing import Tuple

from ..implied_volatility import solve_ivol
from ..numeric_greeks.with_carry import NumericGreeks


def greeks(
        is_european: bool,
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
        n: int
) -> Tuple[float, float, float, float]:
    """A trinomial tree options pricer returning the price and some greeks.

    Args:
        is_european (bool): True for European, false for American.
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        b (float): The cost of carry of the asset.
        v (float): The volatility of the asset.
        n (int): The number of the steps in the tree.

    Returns:
        Tuple[float, float, float, float]: The price, delta, gamma, theta.
    """

    z = 1 if is_call else -1

    dT = T / n
    u = exp(v * sqrt(2 * dT))
    d = exp(-v * sqrt(2 * dT))
    pu = (
        (
            exp(b * dT / 2)
            - exp(-v * sqrt(dT / 2))
        ) / (
            exp(v * sqrt(dT / 2))
            - exp(-v * sqrt(dT / 2))
        )
    ) ** 2
    pd = (
        (
            exp(v * sqrt(dT / 2))
            - exp(b * dT / 2)
        ) / (
            exp(v * sqrt(dT / 2))
            - exp(-v * sqrt(dT / 2))
        )
    ) ** 2
    pm = 1 - pu - pd
    Df = exp(-r * dT)

    option_value = [
        max(0, z * (S * u ** max(i - n, 0) * d ** max(n - i, 0) - K))
        for i in range(1 + 2*n)
    ]

    delta = gamma = theta = nan

    for j in range(n-1, -1, -1):
        for i in range(1 + j*2):

            option_value[i] = (
                pu * option_value[i + 2]
                + pm * option_value[i + 1]
                + pd * option_value[i]
            ) * Df

            if is_european:
                option_value[i] = max(
                    z * (S * u ** max(i - j, 0) * d ** max(j - i, 0) - K),
                    option_value[i]
                )

        if j == 1:
            delta = (option_value[2] - option_value[0]) / (S * u - S * d)
            gamma = (
                (option_value[2] - option_value[1]) / (S * u - S)
                - (option_value[1] - option_value[0]) / (S - S * d)
            ) / (0.5 * (S * u - S * d))
            theta = option_value[1]

    theta = (theta - option_value[0]) / dT / 365

    return option_value[0], delta, gamma, theta


def price(
        is_european: bool,
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
        n: int
) -> float:
    """Calculate the price of an option using a trinomial tree.

    Args:
        is_european (bool): True for European, false for American.
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        b (float): The cost of carry of the asset.
        v (float): The volatility of the asset.
        n (int): The number of the steps in the tree.

    Returns:
        Tuple[float, float, float, float]: The price, delta, gamma, theta.
    """
    p, *_ = greeks(is_european, is_call, S, K, T, r, b, v, n)
    return p


def ivol(
        is_european: bool,
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        p: float,
        n: int,
        *,
        max_iterations: int = 20,
        epsilon=1e-8
) -> float:
    """Calculate the volatility of an option that is implied by the price.

    Args:
        is_european (bool): True for European, false for American.
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to expiry of the option in years.
        r (float): The risk free rate.
        b (float): The cost of carry of the asset.
        p (float): The option price.
        n (int): The number of the steps in the tree.
        max_iterations (int, Optional): The maximum number of iterations before
            a price is returned. Defaults to 20.
        epsilon (float, Optional): The largest acceptable error. Defaults to 1e-8.

    Returns:
        float: The implied volatility.
    """
    return solve_ivol(
        p,
        lambda v: price(is_european, is_call, S, K, T, r, b, v, n),
        max_iterations=max_iterations,
        epsilon=epsilon
    )


def make_numeric_greeks(
        is_european: bool,
        is_call: bool,
        n: int
) -> NumericGreeks:
    """Make a class to generate greeks numerically using finite difference
    methods.

    Args:
        is_european (bool): True for European, false for American.
        is_call (bool): True for a call, false for a put.
        n (int): The number of the steps in the tree.

    Returns:
        NumericGreeks: A class which can generate Greeks using finite difference
            methods.
    """
    # Normalize the price function to match that required by the finite
    # difference methods.
    def evaluate(
            S: float,
            K: float,
            T: float,
            r: float,
            b: float,
            v: float
    ) -> float:
        return price(is_european, is_call, S, K, T, r, b, v, n)

    return NumericGreeks(evaluate)



"""Optional valuation with a European binomial implementation.
"""

from math import comb, exp, log, sqrt

from ..implied_volatility import solve_ivol
from ..numeric_greeks.with_carry import NumericGreeks


def price(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
        n: int
) -> float:
    """A European binomial option pricing tree.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        b (float): The cost of carry of the asset.
        v (float): The volatility of the asset.
        n (int): The number of the steps in the tree.

    Returns:
        float: The price of the option.
    """

    dt = T / n
    u = exp(v * sqrt(dt))
    d = 1 / u
    a = exp(b * dt)
    p = (a - d) / (u - d)
    A = int(log(K / (S * d ** n)) / log(u / d)) + 1

    sum = 0
    if is_call:
        for j in range(A, n+1):
            sum += comb(n, j) * p ** j * (1 - p) ** (n - j) * (
                S * u ** j * d ** (n - j) - K
            )
    else:
        for j in range(A):
            sum += comb(n, j) * p ** j * (1 - p) ** (n - j) * (
                K - S * u ** j * d ** (n - j)
            )

    return exp(-r * T) * sum


def ivol(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        p: float,
        n: int,
        *,
        max_iterations: int = 20,
        epsilon=1e-8
) -> float:
    """Calculate the volatility of an option that is implied by the price.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to expiry of the option in years.
        r (float): The risk free rate.
        b (float): The cost of carry of the asset.
        p (float): The option price.
        n (int): The number of the steps in the tree.
        max_iterations (int, Optional): The maximum number of iterations before
            a price is returned. Defaults to 20.
        epsilon (float, Optional): The largest acceptable error. Defaults to 1e-8.

    Returns:
        float: The implied volatility.
    """
    return solve_ivol(
        p,
        lambda v: price(is_call, S, K, T, r, b, v, n),
        max_iterations=max_iterations,
        epsilon=epsilon
    )


def make_numeric_greeks(is_call: bool, n: int) -> NumericGreeks:
    """_summary_

    Args:
        is_call (bool): True for a call, false for a put.
        n (int): The number of the steps in the tree.

    Returns:
        NumericGreeks: A class which can generate Greeks using finite difference
            methods.
    """
    # Normalize the price function to match that required by the finite
    # difference methods.
    def evaluate(
            S: float,
            K: float,
            T: float,
            r: float,
            b: float,
            v: float
    ) -> float:
        return price(is_call, S, K, T, r, b, v, n)

    return NumericGreeks(evaluate)



"""Tests for distributions"""

from statistics import NormalDist

from jetblack_options.distributions import CHIINV, CND, ND, CNDEV

from .utils import is_close_to

# def test_CHIINV():
#     actual = CHIINV(0.050001, 10)
#     expected = 18.306973
#     assert is_close_to(actual, expected, 1e-6)

def test_cnd():
    actual = CND(1.23564285010596)
    expected = 0.89170432529243
    assert is_close_to(actual, expected, 1e-6)


def test_nd():
    actual = ND(1.23564285010596)
    expected = 0.1859374114061063
    assert is_close_to(actual, expected, 1e-6)

def test_cndev():
    actual = CNDEV(0.7)
    expected = 0.5244005119066525
    assert is_close_to(actual, expected, 1e-6)


def test_cdf():
    nd = NormalDist()
    actual = nd.cdf(1.23564285010596)
    expected = 0.89170432529243
    assert is_close_to(actual, expected, 1e-6)


def test_pdf():
    nd = NormalDist()
    actual = nd.pdf(1.23564285010596)
    expected = 0.1859374114061063
    assert is_close_to(actual, expected, 1e-6)


def test_inv_cdf():
    nd = NormalDist()
    actual = nd.inv_cdf(0.7)
    expected = 0.5244005119066525
    assert is_close_to(actual, expected, 1e-6)



"""Option valuation implementations using the Cox, Ross & Rubinstein
binomial tree.
"""

from math import exp, nan, sqrt
from typing import Tuple

from ..implied_volatility import solve_ivol
from ..numeric_greeks.with_carry import NumericGreeks


def greeks(
        is_european: bool,
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
        n: int
) -> Tuple[float, float, float, float]:
    """A Cox-Ross-Rubinstein binomial tree option pricer returning the price
    and some greeks.

    Args:
        is_european (bool): True for European, false for American.
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        b (float): The cost of carry of the asset.
        v (float): The volatility of the asset.
        n (int): The number of the steps in the tree.

    Returns:
        Tuple[float, float, float, float]: The price, delta, gamma, theta.
    """

    z = 1 if is_call else -1

    dT = T / n
    u = exp(v * sqrt(dT))
    d = 1 / u
    a = exp(b * dT)
    p = (a - d) / (u - d)
    df = exp(-r * dT)

    option_value = [
        max(0, z * (S * u ** i * d ** (n - i) - K))
        for i in range(n+1)
    ]

    delta = gamma = theta = nan

    for j in range(n-1, -1, -1):
        for i in range(j+1):
            if is_european:
                option_value[i] = (
                    p * option_value[i + 1] +
                    (1 - p) * option_value[i]
                ) * df
            else:
                option_value[i] = max(
                    (z * (S * u ** i * d ** (j - i) - K)),
                    (
                        p * option_value[i + 1]
                        + (1 - p) * option_value[i]
                    ) * df
                )

        if j == 2:
            gamma = (
                (option_value[2] - option_value[1]) / (S * u ** 2 - S)
                - (option_value[1] - option_value[0]) / (S - S * d ** 2)
            ) / (0.5 * (S * u ** 2 - S * d ** 2))
            theta = option_value[1]

        if j == 1:
            delta = (option_value[1] - option_value[0]) / (S * u - S * d)

    theta = (theta - option_value[0]) / (2 * dT) / 365

    return option_value[0], delta, gamma, theta


def price(
        is_european: bool,
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
        n: int
) -> float:
    """Calculate the price of an option using a Cox, Ross & Rubenstein
    binomial tree.

    Args:
        is_european (bool): True for European, false for American.
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        b (float): The cost of carry of the asset.
        v (float): The volatility of the asset.
        n (int): The number of the steps in the tree.

    Returns:
        float: The price of the option.
    """
    p, *_ = greeks(is_european, is_call, S, K, T, r, b, v, n)
    return p


def ivol(
        is_european: bool,
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        p: float,
        n: int,
        *,
        max_iterations: int = 20,
        epsilon=1e-8
) -> float:
    """Calculate the volatility of an option that is implied by the price.

    Args:
        is_european (bool): True for European, false for American.
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to expiry of the option in years.
        r (float): The risk free rate.
        b (float): The cost of carry of the asset.
        p (float): The option price.
        n (int): The number of the steps in the tree.
        max_iterations (int, Optional): The maximum number of iterations before
            a price is returned. Defaults to 20.
        epsilon (float, Optional): The largest acceptable error. Defaults to 1e-8.

    Returns:
        float: The implied volatility.
    """
    return solve_ivol(
        p,
        lambda v: price(is_european, is_call, S, K, T, r, b, v, n),
        max_iterations=max_iterations,
        epsilon=epsilon
    )


def make_numeric_greeks(
        is_european: bool,
        is_call: bool,
        n: int
) -> NumericGreeks:
    """Make a class to generate greeks numerically using finite difference
    methods.

    Args:
        is_european (bool): True for European, false for American.
        is_call (bool): True for a call, false for a put.
        n (int): The number of the steps in the tree.

    Returns:
        NumericGreeks: A class which can generate Greeks using finite difference
            methods.
    """
    # Normalize the price function to match that required by the finite
    # difference methods.
    def evaluate(
            S: float,
            K: float,
            T: float,
            r: float,
            b: float,
            v: float
    ) -> float:
        return price(is_european, is_call, S, K, T, r, b, v, n)

    return NumericGreeks(evaluate)



"""Option valuation functions implementing the Bjerksund and Stensland (2002)
American approximation"""

from math import exp, log, sqrt
from statistics import NormalDist
from typing import Callable

from ..distributions import CBND as cbnd
from ..european.generalised_black_scholes import price as bs_price
from ..implied_volatility import solve_ivol
from ..numeric_greeks.with_carry import NumericGreeks

norm = NormalDist()
cdf = norm.cdf


def _phi(
        S: float,
        T: float,
        gamma_: float,
        h: float,
        i: float,
        r: float,
        b: float,
        v: float,
) -> float:
    lambda_ = (-r + gamma_ * b + 0.5 * gamma_ * (gamma_ - 1) * v ** 2) * T
    d = -(log(S / h) + (b + (gamma_ - 0.5) * v ** 2) * T) / (v * sqrt(T))
    kappa = 2 * b / v ** 2 + 2 * gamma_ - 1
    return (
        exp(lambda_) * S ** gamma_ * (
            cdf(d)
            - (i / S) ** kappa * cdf(d - 2 * log(i / S) / (v * sqrt(T))))
    )


def _ksi(
        S: float,
        T2: float,
        gamma_: float,
        h: float,
        I2: float,
        I1: float,
        t1: float,
        r: float,
        b: float,
        v: float,
) -> float:
    e1 = (log(S / I1) + (b + (gamma_ - 0.5) * v ** 2) * t1) / (v * sqrt(t1))
    e2 = (log(I2 ** 2 / (S * I1)) +
          (b + (gamma_ - 0.5) * v ** 2) * t1) / (v * sqrt(t1))
    e3 = (log(S / I1) - (b + (gamma_ - 0.5) * v ** 2) * t1) / (v * sqrt(t1))
    e4 = (log(I2 ** 2 / (S * I1)) -
          (b + (gamma_ - 0.5) * v ** 2) * t1) / (v * sqrt(t1))

    f1 = (log(S / h) + (b + (gamma_ - 0.5) * v ** 2) * T2) / (v * sqrt(T2))
    f2 = (log(I2 ** 2 / (S * h)) +
          (b + (gamma_ - 0.5) * v ** 2) * T2) / (v * sqrt(T2))
    f3 = (log(I1 ** 2 / (S * h)) +
          (b + (gamma_ - 0.5) * v ** 2) * T2) / (v * sqrt(T2))
    f4 = (log(S * I1 ** 2 / (h * I2 ** 2)) +
          (b + (gamma_ - 0.5) * v ** 2) * T2) / (v * sqrt(T2))

    rho = sqrt(t1 / T2)
    lambda_ = -r + gamma_ * b + 0.5 * gamma_ * (gamma_ - 1) * v ** 2
    kappa = 2 * b / (v ** 2) + (2 * gamma_ - 1)

    return (
        exp(lambda_ * T2) *
        S ** gamma_ *
        (
            cbnd(-e1, -f1, rho) -
            (I2 / S) ** kappa * cbnd(-e2, -f2, rho) -
            (I1 / S) ** kappa * cbnd(-e3, -f3, -rho) +
            (I1 / I2) ** kappa * cbnd(-e4, -f4, -rho)
        )
    )


def _call_price(
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
) -> float:

    t1 = 1 / 2 * (sqrt(5) - 1) * T

    if b >= r:
        # Use Black-Scholes as it is never optimal to exercise before maturity.
        return bs_price(True, S, K, T, r, b, v)

    beta = (
        (1 / 2 - b / v ** 2)
        + sqrt((b / v ** 2 - 1 / 2) ** 2 + 2 * r / v ** 2)
    )
    b_infinity = beta / (beta - 1) * K
    b0 = max(K, r / (r - b) * K)

    ht1 = -(b * t1 + 2 * v * sqrt(t1)) * K ** 2 / ((b_infinity - b0) * b0)
    ht2 = -(b * T + 2 * v * sqrt(T)) * K ** 2 / ((b_infinity - b0) * b0)
    I1 = b0 + (b_infinity - b0) * (1 - exp(ht1))
    I2 = b0 + (b_infinity - b0) * (1 - exp(ht2))
    alfa1 = (I1 - K) * I1 ** (-beta)
    alfa2 = (I2 - K) * I2 ** (-beta)

    if S >= I2:
        return S - K
    else:
        return (
            alfa2 * S ** beta
            - alfa2 * _phi(S, t1, beta, I2, I2, r, b, v)
            + _phi(S, t1, 1, I2, I2, r, b, v)
            - _phi(S, t1, 1, I1, I2, r, b, v)
            - K * _phi(S, t1, 0, I2, I2, r, b, v)
            + K * _phi(S, t1, 0, I1, I2, r, b, v)
            + alfa1 * _phi(S, t1, beta, I1, I2, r, b, v)
            - alfa1 * _ksi(S, T, beta, I1, I2, I1, t1, r, b, v)
            + _ksi(S, T, 1, I1, I2, I1, t1, r, b, v)
            - _ksi(S, T, 1, K, I2, I1, t1, r, b, v)
            - K * _ksi(S, T, 0, I1, I2, I1, t1, r, b, v)
            + K * _ksi(S, T, 0, K, I2, I1, t1, r, b, v)
        )


def price(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
) -> float:
    """The Bjerksund and Stensland (2002) American approximation.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        b (float): The cost of carry of the asset.
        v (float): The volatility of the asset.

    Returns:
        float: The price of the option.
    """

    if is_call:
        return _call_price(S, K, T, r, b, v)
    else:
        # Use the Bjerksund and Stensland put-call transformation
        return _call_price(K, S, T, r - b, -b, v)


def ivol(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        p: float,
        *,
        max_iterations: int = 20,
        epsilon=1e-8
) -> float:
    """Calculate the volatility of an option that is implied by the price.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to expiry of the option in years.
        r (float): The risk free rate.
        b (float): The cost of carry of the asset.
        p (float): The option price.
        max_iterations (int, Optional): The maximum number of iterations before
            a price is returned. Defaults to 20.
        epsilon (float, Optional): The largest acceptable error. Defaults to 1e-8.

    Returns:
        float: The implied volatility.
    """
    return solve_ivol(
        p,
        lambda v: price(is_call, S, K, T, r, b, v),
        max_iterations=max_iterations,
        epsilon=epsilon
    )


def make_numeric_greeks(is_call: bool) -> NumericGreeks:
    """Make a class to generate greeks numerically using finite difference
    methods.

    Args:
        is_call (bool): If true the options is a call;  otherwise it is a put.

    Returns:
        NumericGreeks: A class which can generate Greeks using finite difference
            methods.
    """
    # Normalize the price function to match that required by the finite
    # difference methods.
    def evaluate(
            S: float,
            K: float,
            T: float,
            r: float,
            b: float,
            v: float
    ) -> float:
        return price(is_call, S, K, T, r, b, v)

    return NumericGreeks(evaluate)



r"""Black (1976) Options on futures/forwards


* The discounted futures price $ F $,
* Strike price $ K $,
* Risk-free rate $ r $,
* Annual dividend yield $ q $,
* Time to maturity $ \tau = T - t $
* Volatility $ \sigma $.

Most of the formula use one or both of the following terms.

$$
d_1 = \frac{\ln(F/K) + (\sigma^2/2)T}{\sigma\sqrt{T}}
$$

$$
d_2 = \frac{\ln(F/K) - (\sigma^2/2)T}{\sigma\sqrt{T}} = d_1 - \sigma\sqrt{T}
$$

$$
\varphi(x) &= \frac{1}{\sqrt{2\pi}} e^{-\frac{1}{2} x^2}
$$

$$
\Phi(x) &= \frac{1}{\sqrt{2\pi}} \int_{-\infty}^x e^{-\frac{1}{2} y^2} \,dy = 1 - \frac{1}{\sqrt{2\pi}} \int_x^\infty e^{-\frac{1}{2} y^2} \,dy
$$
"""

from math import exp, log, sqrt
from statistics import NormalDist

from ..implied_volatility import solve_ivol
from ..numeric_greeks.without_carry import NumericGreeks

norm = NormalDist()
cdf = norm.cdf
pdf = norm.pdf


def price(
        is_call: bool,
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    r"""Fair value of a futures/forward using Black 76.

    Call price: $ e^{-r \tau}[F\Phi(d_1) - K\Phi(d_2)] $

    Put price: $ e^{-r \tau} [K\Phi(-d_2) -  F\Phi(-d_1)] $

    Args:
        is_call (bool): True for a call, false for a put.
        F (float): The price of the future.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        v (float): The asset volatility.

    Returns:
        float: The option price.
    """

    d1 = (log(F / K) + (v ** 2 / 2) * T) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    if is_call:
        return exp(-r * T) * (F * cdf(d1) - K * cdf(d2))
    else:
        return exp(-r * T) * (K * cdf(-d2) - F * cdf(-d1))


def ivol(
        is_call: bool,
        F: float,
        K: float,
        T: float,
        r: float,
        p: float,
        *,
        max_iterations: int = 20,
        epsilon=1e-8
) -> float:
    r"""Calculate the volatility of a Black 76 option that is implied by the price.

    Args:
        is_call (bool): True for a call, false for a put.
        F (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        p (float): The option price.
        max_iterations (int, Optional): The maximum number of iterations before
            a price is returned. Defaults to 20.
        epsilon (float, Optional): The largest acceptable error. Defaults to 1e-8.

    Returns:
        float: The implied volatility.
    """
    return solve_ivol(
        p,
        lambda v: price(is_call, F, K, T, r, v),
        max_iterations=max_iterations,
        epsilon=epsilon
    )


def make_numeric_greeks(is_call: bool) -> NumericGreeks:
    def evaluate(S: float, K: float, T: float, r: float, b: float) -> float:
        return price(is_call, S, K, T, r, b)
    return NumericGreeks(evaluate)


def delta(
        is_call: bool,
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    r"""The sensitivity of the option to a change in the asset price
    using Black 76.

    Call $\Delta$: $e^{-r \tau} \Phi(d_1)$

    Put $\Delta$: $-e^{-r \tau} \Phi(-d_1)$

    Args:
        is_call (bool): True for a call, false for a put.
        F (float): The current futures price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        v (float): The volatility.

    Returns:
        float: The delta.
    """
    d1 = (log(F / K) + T * (v ** 2 / 2)) / (v * sqrt(T))
    if is_call:
        return exp(-r * T) * cdf(d1)
    else:
        return -exp(-r * T) * cdf(-d1)


def gamma(
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    r"""The second derivative to the change in asset price using Black 76.

    $$
    e^{-r \tau} \frac{\varphi(d_1)}{F\sigma\sqrt{\tau}} = K e^{-r \tau} \frac{\varphi(d_2)}{F^2\sigma\sqrt{\tau}}
    $$

    Args:
        F (float): The current futures price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        v (float): The volatility.

    Returns:
        float: The gamma.
    """
    d1 = (log(F / K) + T * (v ** 2 / 2)) / (v * sqrt(T))

    return exp(-r * T) * pdf(d1) / (F * v * sqrt(T))


def theta(
        is_call: bool,
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    r"""The change in the value of the option with respect to time to expiry
    using Black 76.

    $$
    - \frac{F e^{-r \tau} \varphi(d_1) \sigma}{2 \sqrt{\tau}} - rKe^{-r \tau}\Phi(d_2) + rFe^{-r \tau}\Phi(d_1)
    $$

    $$
    - \frac{F e^{-r \tau} \varphi(d_1) \sigma}{2 \sqrt{\tau}} + rKe^{-r \tau}\Phi(-d_2) - rFe^{-r \tau}\Phi(-d_1)
    $$

    Args:
        is_call (bool): True for a call, false for a put.
        F (float): The current futures price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        v (float): The volatility.

    Returns:
        float: The theta.
    """
    d1 = (log(F / K) + (v ** 2 / 2) * T) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)

    if is_call:
        return (
            -F * exp(-r * T) * pdf(d1) * v / (2 * sqrt(T))
            + r * F * exp(-r * T) * cdf(d1)
            - r * K * exp(-r * T) * cdf(d2)
        )
    else:
        return (
            -F * exp(-r * T) * pdf(d1) * v / (2 * sqrt(T))
            - r * F * exp(-r * T) * cdf(-d1)
            + r * K * exp(-r * T) * cdf(-d2)
        )


def vega(
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    r"""The sensitivity of the options price or a change in the asset volatility
    using Black 76.

    $$
    F e^{-r \tau} \varphi(d_1) \sqrt{\tau} = K e^{-r \tau} \varphi(d_2) \sqrt{\tau}
    $$

    Args:
        F (float): The current futures price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        v (float): The volatility.

    Returns:
        float: The vega.
    """
    d1 = (log(F / K) + (v ** 2 / 2) * T) / (v * sqrt(T))
    return F * exp(-r * T) * pdf(d1) * sqrt(T)


def rho(
        is_call: bool,
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    r"""The sensitivity of the option price to a change in the risk free rate
    using Black 76.

    Call $ \rho = -\tau e^{-r \tau}[F\Phi(d_1) - K\Phi(d_2)] $

    Put $ \rho = -\tau e^{-r \tau} [K\Phi(-d_2) -  F\Phi(-d_1)] $

    Args:
        is_call (bool): True for a call, false for a put.
        F (float): The price of the future.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        v (float): The asset volatility.

    Returns:
        float: The rho.
    """
    d1 = (log(F / K) + (v ** 2 / 2) * T) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    if is_call:
        return -T * exp(-r * T) * (F * cdf(d1) - K * cdf(d2))
    else:
        return -T * exp(-r * T) * (K * cdf(-d2) - F * cdf(-d1))


def vanna(
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    r"""The sensitivity of the option value to the underlying
    asset price and the volatility.

    $$
    \frac{\partial^2 V}{\partial F \partial \sigma} = -e^{-r \tau} \varphi(d_1) \frac{d_2}{\sigma} \, = \frac{\mathcal{V}}{F}\left[1 - \frac{d_1}{\sigma\sqrt{\tau}} \right]
    $$

    Args:
        F (float): The price of the future.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        v (float): The asset volatility.

    Returns:
        float: The vanna.
    """
    d1 = (log(F / K) + T * (v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return -exp(-r * T) * pdf(d1) * d2 / v


def vomma(
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    r"""The second order sensitivity to volatility.

    $$
    F e^{-r \tau} \varphi(d_1) \sqrt{\tau} \frac{d_1 d_2}{\sigma} = \mathcal{V}  \frac{d_1 d_2}{\sigma}
    $$

    Args:
        F (float): The price of the future.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        v (float): The asset volatility.

    Returns:
        float: The vomma
    """
    d1 = (log(F / K) + T * (v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return F * exp(-r * T) * pdf(d1) * sqrt(T) * d1 * d2 / v



"""Garman and Kolhagen (1983) Currency options"""

from math import exp, log, sqrt
from statistics import NormalDist

from ..implied_volatility import solve_ivol
from ..numeric_greeks.with_dividend_yield import NumericGreeks

norm = NormalDist()
cdf = norm.cdf
pdf = norm.pdf
inv_cdf = norm.inv_cdf


def price(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        rf: float,
        v: float,
) -> float:
    """Garman and Kolhagen (1983) Currency options.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The asset price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate of the base currency.
        rf (float): The risk free rate of the quote currency.
        v (float): The asset volatility.

    Returns:
        float: _description_
    """
    # Garman and Kolhagen (1983) Currency options

    d1 = (log(S / K) + (r - rf + v ** 2 / 2) * T) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    if is_call:
        return S * exp(-rf * T) * cdf(d1) - K * exp(-r * T) * cdf(d2)
    else:
        return K * exp(-r * T) * cdf(-d2) - S * exp(-rf * T) * cdf(-d1)


def ivol(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        rf: float,
        p: float,
        *,
        max_iterations: int = 20,
        epsilon=1e-8
) -> float:
    """Calculate the volatility of an option that is implied by the price.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to expiry of the option in years.
        r (float): The risk free rate of the base currency.
        rf (float): The risk free rate of the quote currency.
        p (float): The option price.
        max_iterations (int, Optional): The maximum number of iterations before
            a price is returned. Defaults to 20.
        epsilon (float, Optional): The largest acceptable error. Defaults to 1e-8.

    Returns:
        float: The implied volatility.
    """
    return solve_ivol(
        p,
        lambda v: price(is_call, S, K, T, r, rf, v),
        max_iterations=max_iterations,
        epsilon=epsilon
    )


def make_numeric_greeks(is_call: bool) -> NumericGreeks:
    """Make a class to generate greeks numerically using finite difference
    methods.

    Args:
        is_call (bool): If true the options is a call;  otherwise it is a put.

    Returns:
        NumericGreeks: A class which can generate Greeks using finite difference
            methods.
    """
    # Normalize the price function to match that required by the finite
    # difference methods.
    def evaluate(
            S: float,
            K: float,
            T: float,
            r: float,
            rf: float,
            v: float
    ) -> float:
        return price(is_call, S, K, T, r, rf, v)

    return NumericGreeks(evaluate)



"""Black Scholes variance analytic solutions"""

from math import exp, log, sqrt
from statistics import NormalDist

norm = NormalDist()
cdf = norm.cdf


def price(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
) -> float:
    # The generalized Black and Scholes formula on variance form

    d1 = (log(S / K) + (b + v / 2) * T) / sqrt(v * T)
    d2 = d1 - sqrt(v * T)

    if is_call:
        return S * exp((b - r) * T) * cdf(d1) - K * exp(-r * T) * cdf(d2)
    else:
        return K * exp(-r * T) * cdf(-d2) - S * exp((b - r) * T) * cdf(-d1)
